/* BdfServer - Copyright (c) 2010-2022 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.bdfserver.tools.instruction;

import fr.exemole.bdfserver.api.BdfServer;
import fr.exemole.bdfserver.api.balayage.Balayage;
import fr.exemole.bdfserver.api.instruction.BdfCommandParameters;
import fr.exemole.bdfserver.api.instruction.BdfInstruction;
import fr.exemole.bdfserver.api.instruction.BdfInstructionConstants;
import fr.exemole.bdfserver.api.instruction.BdfParameters;
import fr.exemole.bdfserver.api.instruction.OutputParameters;
import fr.exemole.bdfserver.api.instruction.TypedBdfInstruction;
import fr.exemole.bdfserver.api.interaction.Domain;
import fr.exemole.bdfserver.api.interaction.Domains;
import fr.exemole.bdfserver.api.interaction.InteractionConstants;
import fr.exemole.bdfserver.api.managers.AccessManager;
import fr.exemole.bdfserver.api.managers.BalayageManager;
import fr.exemole.bdfserver.api.managers.PermissionManager;
import fr.exemole.bdfserver.api.managers.ScrutariExportManager;
import fr.exemole.bdfserver.api.managers.SelectionManager;
import fr.exemole.bdfserver.api.managers.SqlExportManager;
import fr.exemole.bdfserver.api.managers.TableExportManager;
import fr.exemole.bdfserver.api.managers.TransformationManager;
import fr.exemole.bdfserver.api.roles.Role;
import fr.exemole.bdfserver.api.users.BdfUser;
import fr.exemole.bdfserver.tools.configuration.ConfigurationUtils;
import fr.exemole.bdfserver.tools.exportation.table.CellEngineUtils;
import fr.exemole.bdfserver.tools.exportation.transformation.BdfTransformationUtils;
import java.text.ParseException;
import java.util.Comparator;
import net.fichotheque.Fichotheque;
import net.fichotheque.Subset;
import net.fichotheque.SubsetItem;
import net.fichotheque.SubsetKey;
import net.fichotheque.addenda.Addenda;
import net.fichotheque.addenda.Document;
import net.fichotheque.album.Album;
import net.fichotheque.album.Illustration;
import net.fichotheque.album.metadata.AlbumDim;
import net.fichotheque.corpus.Corpus;
import net.fichotheque.corpus.FicheMeta;
import net.fichotheque.exportation.access.AccessDef;
import net.fichotheque.exportation.balayage.BalayageContentDescription;
import net.fichotheque.exportation.balayage.BalayageDescription;
import net.fichotheque.exportation.scrutari.ScrutariExportDef;
import net.fichotheque.exportation.sql.SqlExportDef;
import net.fichotheque.exportation.table.CellEngine;
import net.fichotheque.exportation.table.TableExportContentDescription;
import net.fichotheque.exportation.table.TableExportDescription;
import net.fichotheque.exportation.transformation.TemplateContentDescription;
import net.fichotheque.exportation.transformation.TemplateDescription;
import net.fichotheque.exportation.transformation.TemplateKey;
import net.fichotheque.exportation.transformation.TransformationDescription;
import net.fichotheque.exportation.transformation.TransformationKey;
import net.fichotheque.extraction.ExtractionContext;
import net.fichotheque.selection.FicheQuery;
import net.fichotheque.selection.FichothequeQueries;
import net.fichotheque.selection.MotcleQuery;
import net.fichotheque.selection.SelectionDef;
import net.fichotheque.sphere.Redacteur;
import net.fichotheque.sphere.Sphere;
import net.fichotheque.thesaurus.Motcle;
import net.fichotheque.thesaurus.Thesaurus;
import net.fichotheque.tools.parsers.TypoParser;
import net.fichotheque.tools.selection.FichothequeQueriesBuilder;
import net.fichotheque.tools.selection.SelectionDOMUtils;
import net.fichotheque.utils.BalayageUtils;
import net.fichotheque.utils.Comparators;
import net.fichotheque.utils.FichothequeUtils;
import net.fichotheque.utils.SelectionUtils;
import net.fichotheque.utils.TableExportUtils;
import net.fichotheque.utils.selection.FicheSelectorBuilder;
import net.fichotheque.utils.selection.MotcleSelectorBuilder;
import net.mapeadores.util.annotation.Nullable;
import net.mapeadores.util.io.ResourceUtils;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.localisation.LangPreference;
import net.mapeadores.util.localisation.LangPreferenceBuilder;
import net.mapeadores.util.localisation.MessageLocalisation;
import net.mapeadores.util.localisation.UserLangContext;
import net.mapeadores.util.logging.ErrorMessageException;
import net.mapeadores.util.request.RequestMap;
import net.mapeadores.util.request.RequestUtils;
import net.mapeadores.util.servlets.ResponseHandler;
import net.mapeadores.util.text.LabelChange;
import net.mapeadores.util.text.LabelChangeBuilder;
import net.mapeadores.util.text.RelativePath;
import net.mapeadores.util.text.StringUtils;
import net.mapeadores.util.text.TypoOptions;
import net.mapeadores.util.xml.DOMUtils;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;


/**
 *
 * @author Vincent Calame
 */
public final class BdfInstructionUtils {

    public final static BdfInstruction NOTFOUND_BDFINSTRUCTION = new NotFoundBdfInstruction();


    private BdfInstructionUtils() {
    }

    public static MessageLocalisation getMessageLocalisation(RequestMap requestMap, BdfServer bdfServer, @Nullable UserLangContext userLangContext) {
        if (userLangContext != null) {
            return bdfServer.getL10nManager().getMessageLocalisation(userLangContext);
        } else {
            return bdfServer.getL10nManager().getMessageLocalisation(getLangPreference(requestMap, bdfServer));
        }
    }

    public static LangPreference getLangPreference(RequestMap requestMap, BdfServer bdfServer) {
        LangPreferenceBuilder langPreferenceBuilder = new LangPreferenceBuilder();
        RequestUtils.checkLangPreference(requestMap, langPreferenceBuilder);
        ConfigurationUtils.checkLangPreference(bdfServer.getLangConfiguration(), langPreferenceBuilder);
        langPreferenceBuilder.addLang(Lang.build("fr"));
        return langPreferenceBuilder.toLangPreference();
    }

    public static SubsetItem checkMasterSubsetItem(Corpus corpus, int id, BdfUser bdfUser) throws ErrorMessageException {
        Subset masterSubset = corpus.getMasterSubset();
        if (masterSubset != null) {
            SubsetItem masterSubsetItem = masterSubset.getSubsetItemById(id);
            if (masterSubsetItem == null) {
                throw BdfErrors.error("_ error.unknown.mastersubsetitem", id, masterSubset.getSubsetKeyString());
            }
            return masterSubsetItem;
        } else {
            return null;
        }
    }

    public static String getMandatoryParameter(RequestMap requestMap, String paramName) throws ErrorMessageException {
        String paramValue = requestMap.getParameter(paramName);
        if (paramValue == null) {
            throw BdfErrors.emptyMandatoryParameter(paramName);
        }
        return paramValue;
    }

    public static String[] getMandatoryParameterValues(RequestMap requestMap, String paramName) throws ErrorMessageException {
        String[] paramValues = requestMap.getParameterValues(paramName);
        if (paramValues == null) {
            throw BdfErrors.emptyMandatoryParameter(paramName);
        }
        return paramValues;
    }

    public static int getMandatoryIntegerParameter(RequestMap requestMap, String paramName) throws ErrorMessageException {
        String paramValue = requestMap.getParameter(paramName);
        if (paramValue == null) {
            throw BdfErrors.emptyMandatoryParameter(paramName);
        }
        try {
            int value = Integer.parseInt(paramValue);
            return value;
        } catch (NumberFormatException nfe) {
            throw BdfErrors.wrongParameterValue(paramName, paramValue);
        }
    }

    public static Corpus getMandatoryCorpus(Fichotheque fichotheque, RequestMap requestMap) throws ErrorMessageException {
        return (Corpus) getMandatorySubset(fichotheque, requestMap, SubsetKey.CATEGORY_CORPUS);
    }

    public static Thesaurus getMandatoryThesaurus(Fichotheque fichotheque, RequestMap requestMap) throws ErrorMessageException {
        return (Thesaurus) getMandatorySubset(fichotheque, requestMap, SubsetKey.CATEGORY_THESAURUS);
    }

    public static Sphere getMandatorySphere(Fichotheque fichotheque, RequestMap requestMap) throws ErrorMessageException {
        return (Sphere) getMandatorySubset(fichotheque, requestMap, SubsetKey.CATEGORY_SPHERE);
    }

    public static Album getMandatoryAlbum(Fichotheque fichotheque, RequestMap requestMap) throws ErrorMessageException {
        return (Album) getMandatorySubset(fichotheque, requestMap, SubsetKey.CATEGORY_ALBUM);
    }

    public static Addenda getMandatoryAddenda(Fichotheque fichotheque, RequestMap requestMap) throws ErrorMessageException {
        return (Addenda) getMandatorySubset(fichotheque, requestMap, SubsetKey.CATEGORY_ADDENDA);
    }

    public static Subset getMandatorySubset(Fichotheque fichotheque, RequestMap requestMap, short subsetCategory) throws ErrorMessageException {
        return getMandatorySubset(fichotheque, requestMap, subsetCategory, SubsetKey.categoryToString(subsetCategory));
    }

    public static Subset getMandatorySubset(Fichotheque fichotheque, RequestMap requestMap, short subsetCategory, String paramName) throws ErrorMessageException {
        String paramValue = getMandatoryParameter(requestMap, paramName);
        try {
            SubsetKey subsetKey = SubsetKey.parse(subsetCategory, paramValue);
            Subset subset = fichotheque.getSubset(subsetKey);
            if (subset == null) {
                throw BdfErrors.unknownParameterValue(paramName, paramValue);
            }
            return subset;
        } catch (ParseException pe) {
            throw BdfErrors.wrongParameterValue(paramName, paramValue);
        }
    }

    public static Subset getMandatorySubset(Fichotheque fichotheque, RequestMap requestMap) throws ErrorMessageException {
        return getMandatorySubset(fichotheque, requestMap, InteractionConstants.SUBSET_PARAMNAME);
    }

    public static Subset getMandatorySubset(Fichotheque fichotheque, RequestMap requestMap, String paramName) throws ErrorMessageException {
        String paramValue = getMandatoryParameter(requestMap, paramName);
        try {
            SubsetKey subsetKey = SubsetKey.parse(paramValue);
            Subset subset = fichotheque.getSubset(subsetKey);
            if (subset == null) {
                throw BdfErrors.unknownParameterValue(paramName, paramValue);
            }
            return subset;
        } catch (ParseException pe) {
            throw BdfErrors.wrongParameterValue(paramName, paramValue);
        }
    }

    public static AlbumDim getMandatoryAlbumDim(Fichotheque fichotheque, RequestMap requestMap) throws ErrorMessageException {
        return getMandatoryAlbumDim(fichotheque, requestMap, "albumdim");
    }

    public static AlbumDim getMandatoryAlbumDim(Fichotheque fichotheque, RequestMap requestMap, String paramName) throws ErrorMessageException {
        Album album = getMandatoryAlbum(fichotheque, requestMap);
        return getMandatoryAlbumDim(fichotheque, requestMap, album, paramName);
    }

    public static AlbumDim getMandatoryAlbumDim(Fichotheque fichotheque, RequestMap requestMap, Album album, String paramName) throws ErrorMessageException {
        String paramValue = getMandatoryParameter(requestMap, paramName);
        AlbumDim albumDim = album.getAlbumMetadata().getAlbumDimByName(paramValue);
        if (albumDim == null) {
            throw BdfErrors.unknownParameterValue(paramName, paramValue);
        }
        return albumDim;
    }

    public static FicheMeta getMandatoryFicheMeta(Fichotheque fichotheque, RequestMap requestMap, BdfUser bdfUser) throws ErrorMessageException {
        Corpus corpus = getMandatoryCorpus(fichotheque, requestMap);
        int id = BdfInstructionUtils.getMandatoryId(requestMap);
        FicheMeta ficheMeta = corpus.getFicheMetaById(id);
        if (ficheMeta == null) {
            String corpusTitle = FichothequeUtils.getTitle(corpus, bdfUser.getWorkingLang());
            throw BdfErrors.error("_ error.unknown.fiche", corpusTitle, id);
        }
        return ficheMeta;
    }

    public static FicheMeta getMandatoryFicheMeta(Fichotheque fichotheque, RequestMap requestMap) throws ErrorMessageException {
        return (FicheMeta) getMandatorySubsetItem(fichotheque, requestMap, SubsetKey.CATEGORY_CORPUS);
    }

    public static Motcle getMandatoryMotcle(Fichotheque fichotheque, RequestMap requestMap) throws ErrorMessageException {
        String idString = requestMap.getParameter(InteractionConstants.IDMTCL_PARAMNAME);
        if (idString == null) {
            return (Motcle) getMandatorySubsetItem(fichotheque, requestMap, SubsetKey.CATEGORY_THESAURUS);
        }
        return getMandatoryMotcle(fichotheque, requestMap, InteractionConstants.IDMTCL_PARAMNAME);
    }

    public static Motcle getMandatoryMotcle(Fichotheque fichotheque, RequestMap requestMap, String paramName) throws ErrorMessageException {
        String idString = requestMap.getParameter(paramName);
        Thesaurus thesaurus = getMandatoryThesaurus(fichotheque, requestMap);
        Motcle motcle = null;
        try {
            motcle = getMotcle(thesaurus, idString);
        } catch (NumberFormatException nfe) {
            throw BdfErrors.wrongParameterValue(paramName, idString);
        }
        if (motcle == null) {
            throw BdfErrors.unknownParameterValue(paramName, idString);
        }
        return motcle;
    }

    public static Motcle getOptionnalMotcle(RequestMap requestMap, Thesaurus thesaurus) throws ErrorMessageException {
        String idmtclValue = requestMap.getParameter(InteractionConstants.IDMTCL_PARAMNAME);
        if (idmtclValue == null) {
            return (Motcle) getOptionnalSubsetItem(requestMap, thesaurus);
        }
        try {
            Motcle motcle = getMotcle(thesaurus, idmtclValue);
            return motcle;
        } catch (NumberFormatException nfe) {
            throw BdfErrors.wrongParameterValue(InteractionConstants.IDMTCL_PARAMNAME, idmtclValue);
        }
    }

    public static Motcle getMotcle(Thesaurus thesaurus, String idString) throws NumberFormatException {
        if (thesaurus.isIdalphaType()) {
            Motcle motcle = thesaurus.getMotcleByIdalpha(idString);
            if (motcle != null) {
                return motcle;
            } else {
                try {
                    int motcleid = Integer.parseInt(idString);
                    return thesaurus.getMotcleById(motcleid);
                } catch (NumberFormatException nfe) {
                    return null;
                }
            }
        } else {
            int motcleid = Integer.parseInt(idString);
            return thesaurus.getMotcleById(motcleid);
        }
    }

    public static Redacteur getMandatoryRedacteur(Fichotheque fichotheque, RequestMap requestMap) throws ErrorMessageException {
        return (Redacteur) getMandatorySubsetItem(fichotheque, requestMap, SubsetKey.CATEGORY_SPHERE);
    }

    public static Document getMandatoryDocument(Fichotheque fichotheque, RequestMap requestMap) throws ErrorMessageException {
        return (Document) getMandatorySubsetItem(fichotheque, requestMap, SubsetKey.CATEGORY_ADDENDA);
    }

    public static Illustration getMandatoryIllustration(Fichotheque fichotheque, RequestMap requestMap) throws ErrorMessageException {
        return (Illustration) getMandatorySubsetItem(fichotheque, requestMap, SubsetKey.CATEGORY_ALBUM);
    }

    public static Illustration getMandatoryIllustration(RequestMap requestMap, Album album) throws ErrorMessageException {
        return (Illustration) getMandatorySubsetItem(requestMap, album);
    }

    public static LabelChange getLabelChange(RequestMap requestMap, String paramPrefix, boolean withTypoParsing) {
        LabelChangeBuilder result = new LabelChangeBuilder();
        int prefixLength = paramPrefix.length();
        for (String paramName : requestMap.getParameterNameSet()) {
            if (!paramName.startsWith(paramPrefix)) {
                continue;
            }
            try {
                Lang currentLang = Lang.parse(paramName.substring(prefixLength));
                String newLabelString = requestMap.getParameter(paramName);
                if (withTypoParsing) {
                    TypoOptions typoOptions = TypoOptions.getTypoOptions(currentLang.toLocale());
                    newLabelString = TypoParser.parseTypo(newLabelString, typoOptions);
                }
                result.putLabel(currentLang, newLabelString);

            } catch (ParseException pe) {
            }
        }
        return result.toLabelChange();
    }

    private static SubsetItem getMandatorySubsetItem(Fichotheque fichotheque, RequestMap requestMap, short subsetCategory) throws ErrorMessageException {
        int id = BdfInstructionUtils.getMandatoryId(requestMap);
        Subset subset = getMandatorySubset(fichotheque, requestMap, subsetCategory);
        SubsetItem subsetItem = subset.getSubsetItemById(id);
        if (subsetItem == null) {
            throw BdfErrors.unknownParameterValue(InteractionConstants.ID_PARAMNAME, String.valueOf(id));
        }
        return subsetItem;
    }

    private static SubsetItem getMandatorySubsetItem(RequestMap requestMap, Subset subset) throws ErrorMessageException {
        int id = BdfInstructionUtils.getMandatoryId(requestMap);
        SubsetItem subsetItem = subset.getSubsetItemById(id);
        if (subsetItem == null) {
            throw BdfErrors.unknownParameterValue(InteractionConstants.ID_PARAMNAME, String.valueOf(id));
        }
        return subsetItem;
    }

    public static int getMandatoryId(RequestMap requestMap) throws ErrorMessageException {
        int id = getMandatoryIntegerParameter(requestMap, InteractionConstants.ID_PARAMNAME);
        if (id < 1) {
            throw BdfErrors.wrongParameterValue(InteractionConstants.ID_PARAMNAME, requestMap.getParameter(InteractionConstants.ID_PARAMNAME));
        }
        return id;
    }

    public static SubsetItem getOptionnalSubsetItem(RequestMap requestMap, Subset subset) throws ErrorMessageException {
        return getOptionnalSubsetItem(requestMap, subset, InteractionConstants.ID_PARAMNAME);
    }

    public static SubsetItem getOptionnalSubsetItem(RequestMap requestMap, Subset subset, String paramName) throws ErrorMessageException {
        String idString = requestMap.getParameter(paramName);
        if (idString == null) {
            return null;
        }
        int id = 0;
        try {
            id = Integer.parseInt(idString);
        } catch (NumberFormatException nfe) {
            throw BdfErrors.wrongParameterValue(paramName, idString);
        }
        if (id < 1) {
            throw BdfErrors.wrongParameterValue(paramName, idString);
        }
        SubsetItem subsetItem = subset.getSubsetItemById(id);
        return subsetItem;
    }

    public static Role getMandatoryRole(PermissionManager permissionManager, RequestMap requestMap) throws ErrorMessageException {
        String roleString = getMandatoryParameter(requestMap, InteractionConstants.ROLE_PARAMNAME);
        Role role = permissionManager.getRole(roleString);
        if (role == null) {
            throw BdfErrors.unknownParameterValue(InteractionConstants.ROLE_PARAMNAME, roleString);
        }
        return role;
    }

    public static ScrutariExportDef getMandatoryScrutariExportDef(ScrutariExportManager scrutariExportManager, RequestMap requestMap) throws ErrorMessageException {
        String scrutariexportString = getMandatoryParameter(requestMap, InteractionConstants.SCRUTARIEXPORT_PARAMNAME);
        ScrutariExportDef scrutariExportDef = scrutariExportManager.getScrutariExportDef(scrutariexportString);
        if (scrutariExportDef == null) {
            throw BdfErrors.unknownParameterValue(InteractionConstants.SCRUTARIEXPORT_PARAMNAME, scrutariexportString);
        }
        return scrutariExportDef;
    }

    public static SelectionDef getMandatorySelectionDef(SelectionManager selectionManager, RequestMap requestMap) throws ErrorMessageException {
        String param_selection = getMandatoryParameter(requestMap, InteractionConstants.SELECTION_PARAMNAME);
        SelectionDef selectionDef = selectionManager.getSelectionDef(param_selection);
        if (selectionDef == null) {
            throw BdfErrors.unknownParameterValue(InteractionConstants.SELECTION_PARAMNAME, param_selection);
        }
        return selectionDef;
    }

    public static SqlExportDef getMandatorySqlExportDef(SqlExportManager sqlExportManager, RequestMap requestMap) throws ErrorMessageException {
        String param_sqlexport = getMandatoryParameter(requestMap, InteractionConstants.SQLEXPORT_PARAMNAME);
        SqlExportDef sqlExportDef = sqlExportManager.getSqlExportDef(param_sqlexport);
        if (sqlExportDef == null) {
            throw BdfErrors.unknownParameterValue(InteractionConstants.SQLEXPORT_PARAMNAME, param_sqlexport);
        }
        return sqlExportDef;
    }

    public static AccessDef getMandatoryAccessDef(AccessManager accessManager, RequestMap requestMap) throws ErrorMessageException {
        String param_access = getMandatoryParameter(requestMap, InteractionConstants.ACCESS_PARAMNAME);
        AccessDef accessDef = accessManager.getAccessDef(param_access);
        if (accessDef == null) {
            throw BdfErrors.unknownParameterValue(InteractionConstants.ACCESS_PARAMNAME, param_access);
        }
        return accessDef;
    }

    public static Balayage getMandatoryBalayage(BalayageManager balayageManager, RequestMap requestMap) throws ErrorMessageException {
        String param_balayage = getMandatoryParameter(requestMap, InteractionConstants.BALAYAGE_PARAMNAME);
        Balayage balayage = balayageManager.getBalayage(param_balayage);
        if (balayage == null) {
            throw BdfErrors.unknownParameterValue(InteractionConstants.BALAYAGE_PARAMNAME, param_balayage);
        }
        return balayage;
    }

    public static RelativePath getMandatoryResourcePath(RequestMap requestMap) throws ErrorMessageException {
        String path = getMandatoryParameter(requestMap, InteractionConstants.PATH_PARAMNAME);
        RelativePath relativePath;
        try {
            relativePath = RelativePath.parse(path);
            if (!ResourceUtils.isValidResourceRelativePath(relativePath)) {
                relativePath = null;
            }
        } catch (ParseException pe) {
            relativePath = null;
        }
        if (relativePath == null) {
            throw BdfErrors.wrongParameterValue(InteractionConstants.PATH_PARAMNAME, path);
        }
        return relativePath;
    }

    public static String getMode(RequestMap requestMap) throws ErrorMessageException {
        String mode = requestMap.getParameter(InteractionConstants.MODE_PARAMNAME);
        if (mode == null) {
            return "";
        }
        mode = mode.trim();
        if (mode.length() > 0) {
            try {
                StringUtils.checkTechnicalName(mode, false);
            } catch (ParseException pe) {
                throw BdfErrors.error("_ error.wrong.mode", mode);
            }
        }
        return mode;
    }

    public static boolean checkConfirmation(RequestMap requestMap, BdfServer bdfServer) {
        String checkConfirmation = requestMap.getParameter(InteractionConstants.CONFIRMATIONCHECK_PARAMNAME);
        if (checkConfirmation == null) {
            return false;
        }
        return (checkConfirmation.equals(InteractionConstants.OK_CONFIRMATIONCHECK_PARAMVALUE));
    }

    public static String getFicheGetLink(FicheMeta ficheMeta, String extension) {
        return "fiches/" + ficheMeta.getSubsetName() + "-" + String.valueOf(ficheMeta.getId()) + "." + extension;
    }

    public static String getFicheGetLink(FicheMeta ficheMeta, String extension, String name) {
        return "fiches/" + ficheMeta.getSubsetName() + "-" + String.valueOf(ficheMeta.getId()) + "-" + name + "." + extension;
    }

    public static String getCompilationGetLink(String extension) {
        return "fiches/compilation." + extension;
    }

    public static String getCompilationGetLink(String extension, String name) {
        return "fiches/_compilation-" + name + "." + extension;
    }

    public static BalayageDescription getMandatoryBalayageDescription(BalayageManager manager, RequestMap requestMap) throws ErrorMessageException {
        String balayageName = getMandatoryParameter(requestMap, InteractionConstants.BALAYAGE_PARAMNAME);
        for (BalayageDescription balayageDescription : manager.getBalayageDescriptionList()) {
            if (balayageDescription.getName().equals(balayageName)) {
                return balayageDescription;
            }
        }
        throw BdfErrors.unknownParameterValue(InteractionConstants.BALAYAGE_PARAMNAME, balayageName);
    }

    public static BalayageContentDescription getMandatoryBalayageContentDescription(BalayageDescription balayageDescription, RequestMap requestMap) throws ErrorMessageException {
        String contentPath = BdfInstructionUtils.getMandatoryPath(requestMap);
        BalayageContentDescription description = BalayageUtils.getBalayageContentDescription(balayageDescription, contentPath);
        if (description == null) {
            throw BdfErrors.unknownParameterValue(InteractionConstants.PATH_PARAMNAME, contentPath);
        }
        return description;
    }

    public static TableExportDescription getMandatoryTableExportDescription(TableExportManager manager, RequestMap requestMap) throws ErrorMessageException {
        String tableExportName = getMandatoryParameter(requestMap, InteractionConstants.TABLEEXPORT_PARAMNAME);
        for (TableExportDescription tableExportDescription : manager.getTableExportDescriptionList()) {
            if (tableExportDescription.getName().equals(tableExportName)) {
                return tableExportDescription;
            }
        }
        throw BdfErrors.unknownParameterValue(InteractionConstants.TABLEEXPORT_PARAMNAME, tableExportName);
    }

    public static TableExportContentDescription getMandatoryTableExportContentDescription(TableExportDescription tableExportDescription, RequestMap requestMap) throws ErrorMessageException {
        String contentPath = BdfInstructionUtils.getMandatoryPath(requestMap);
        TableExportContentDescription description = TableExportUtils.getTableExportContentDescription(tableExportDescription, contentPath);
        if (description == null) {
            throw BdfErrors.unknownParameterValue(InteractionConstants.PATH_PARAMNAME, contentPath);
        }
        return description;
    }

    public static String getMandatoryPath(RequestMap requestMap) throws ErrorMessageException {
        String s = getMandatoryParameter(requestMap, InteractionConstants.PATH_PARAMNAME);
        String path = s.trim();
        if (path.length() == 0) {
            throw BdfErrors.wrongParameterValue(InteractionConstants.PATH_PARAMNAME, s);
        }
        return path;
    }

    public static TransformationDescription getMandatoryTransformationDescription(BdfServer bdfServer, RequestMap requestMap) throws ErrorMessageException {
        TransformationManager transformationManager = bdfServer.getTransformationManager();
        String s = getMandatoryParameter(requestMap, InteractionConstants.TRANSFORMATION_PARAMNAME);
        try {
            TransformationKey transformationKey = TransformationKey.parse(s);
            return transformationManager.getTransformationDescription(transformationKey);
        } catch (ParseException pe) {
            throw BdfErrors.wrongParameterValue(InteractionConstants.TRANSFORMATION_PARAMNAME, s);
        }
    }

    public static TemplateDescription getMandatoryTemplateDescription(BdfServer bdfServer, RequestMap requestMap) throws ErrorMessageException {
        TransformationManager transformationManager = bdfServer.getTransformationManager();
        String templateKeyString = getMandatoryParameter(requestMap, InteractionConstants.TEMPLATE_PARAMNAME);
        TemplateKey templateKey;
        try {
            templateKey = TemplateKey.parse(templateKeyString);
        } catch (ParseException pe) {
            throw BdfErrors.unknownParameterValue(InteractionConstants.TEMPLATE_PARAMNAME, templateKeyString);
        }
        TemplateDescription templateDescription = BdfTransformationUtils.getTemplateDescription(transformationManager, templateKey);
        if (templateDescription == null) {
            throw BdfErrors.unknownParameterValue(InteractionConstants.TEMPLATE_PARAMNAME, templateKeyString);
        } else {
            return templateDescription;
        }
    }

    public static TemplateContentDescription getMandatoryTemplateContentDescription(TemplateDescription templateDescription, RequestMap requestMap) throws ErrorMessageException {
        String contentPath = BdfInstructionUtils.getMandatoryPath(requestMap);
        TemplateContentDescription description = templateDescription.getTemplateContentDescription(contentPath);
        if (description == null) {
            throw BdfErrors.unknownParameterValue(InteractionConstants.PATH_PARAMNAME, contentPath);
        }
        return description;
    }

    public static String getExtensionDomain(String registrationName) {
        return Domain.getCompleteDomain(Domains.EXTENSION, registrationName);
    }

    public static boolean ownsToExtension(BdfCommandParameters bdfCommandParameters, String registrationName) {
        return isExtensionDomain(bdfCommandParameters.getDomain(), registrationName);
    }

    public static boolean ownsToExtension(OutputParameters parameters, String registrationName) {
        return isExtensionDomain(parameters.getDomain(), registrationName);
    }

    public static boolean isExtensionDomain(Domain domain, String registrationName) {
        return (domain.getFirstPart().equals(Domains.EXTENSION)) && (domain.getSecondPart().equals(registrationName));
    }

    public static String getAbsolutePageName(String domain, String page) {
        return domain + ":" + page;
    }

    public static String getAbsoluteCommandName(String domain, String commandName) {
        return domain + ":" + commandName;
    }

    public static String getPageOutput(String pageName) {
        return "page|" + pageName;
    }

    public static String getStreamOutput(String streamName) {
        return "stream|" + streamName;
    }

    public static String getJsonOutput(String jsonName) {
        return "json|" + jsonName;
    }

    public static boolean isJsonOutput(BdfInstruction bdfInstruction) {
        if (bdfInstruction instanceof TypedBdfInstruction) {
            return (((TypedBdfInstruction) bdfInstruction).getOutputType() == TypedBdfInstruction.JSON_TYPE);
        }
        return false;
    }

    public static String getPhraseParamPrefix(String phraseName) {
        return "phrase_" + phraseName + "/";
    }

    public static CellEngine getCellEngine(BdfParameters bdfParameters, RequestMap requestMap) {
        BdfServer bdfServer = bdfParameters.getBdfServer();
        String tableExportName = requestMap.getParameter(InteractionConstants.TABLEEXPORT_PARAMNAME);
        if ((tableExportName == null) || (tableExportName.isEmpty())) {
            return TableExportUtils.EMPTY_CELLENGINE;
        }
        ExtractionContext extractionContext = bdfParameters.getDefaultExtractionContext();
        CellEngine cellEngine = CellEngineUtils.newCellEngine(bdfServer, extractionContext, tableExportName);
        if (cellEngine == null) {
            return TableExportUtils.EMPTY_CELLENGINE;
        } else {
            return cellEngine;
        }
    }

    public static Comparator<FicheMeta> getComparator(Lang lang, RequestMap requestMap) throws ErrorMessageException {
        String sort = requestMap.getParameter(InteractionConstants.SORT_PARAMNAME);
        if ((sort == null) || (sort.isEmpty())) {
            return Comparators.FICHEID_ASC;
        }
        try {
            sort = SelectionUtils.checkSortType(sort);
            return Comparators.getComparator(sort, lang);
        } catch (IllegalArgumentException iae) {
            throw BdfErrors.error("_ error.unknown.parametervalue", InteractionConstants.SORT_PARAMNAME, sort);
        }
    }

    public static void populateFromXml(BdfServer bdfServer, RequestMap requestMap, FicheSelectorBuilder builder) throws ErrorMessageException {
        String xml = BdfInstructionUtils.getMandatoryParameter(requestMap, InteractionConstants.XML_PARAMNAME);
        org.w3c.dom.Document document;
        try {
            document = DOMUtils.parseDocument(xml);
        } catch (SAXException sae) {
            throw BdfErrors.error("_ error.exception.xml.sax", sae.getMessage());
        }
        Element root = document.getDocumentElement();
        switch (root.getTagName()) {
            case "fiche-query":
            case "fiche-select":
                FicheQuery ficheQuery = SelectionDOMUtils.getFicheConditionEntry(bdfServer.getFichotheque(), root).getFicheQuery();
                builder.add(ficheQuery);
                break;
            default:
                FichothequeQueriesBuilder fichothequeQueriesBuilder = new FichothequeQueriesBuilder();
                DOMUtils.readChildren(root, SelectionDOMUtils.getQueryElementConsumer(bdfServer.getFichotheque(), fichothequeQueriesBuilder));
                FichothequeQueries fichothequeQueries = fichothequeQueriesBuilder.toFichothequeQueries();
                if (fichothequeQueries.getFicheQueryList().isEmpty()) {
                    throw BdfErrors.unsupportedParameterValue(InteractionConstants.XML_PARAMNAME, "Missing fiche-query elements");
                }
                for (FicheQuery fq : fichothequeQueries.getFicheQueryList()) {
                    builder.add(fq);
                }
        }
    }

    public static void populateFromSelection(BdfServer bdfServer, RequestMap requestMap, FicheSelectorBuilder builder) throws ErrorMessageException {
        SelectionDef selectionDef = BdfInstructionUtils.getMandatorySelectionDef(bdfServer.getSelectionManager(), requestMap);
        for (FicheQuery ficheQuery : selectionDef.getFichothequeQueries().getFicheQueryList()) {
            builder.add(ficheQuery);
        }
    }

    public static void populateFromXml(BdfServer bdfServer, RequestMap requestMap, MotcleSelectorBuilder builder) throws ErrorMessageException {
        String xml = BdfInstructionUtils.getMandatoryParameter(requestMap, InteractionConstants.XML_PARAMNAME);
        org.w3c.dom.Document document;
        try {
            document = DOMUtils.parseDocument(xml);
        } catch (SAXException sae) {
            throw BdfErrors.error("_ error.exception.xml.sax", sae.getMessage());
        }
        Element root = document.getDocumentElement();
        switch (root.getTagName()) {
            case "motcle-query":
            case "motcle-select":
                MotcleQuery motcleQuery = SelectionDOMUtils.getMotcleConditionEntry(bdfServer.getFichotheque(), root).getMotcleQuery();
                builder.add(motcleQuery);
                break;
            default:
                FichothequeQueriesBuilder fichothequeQueriesBuilder = new FichothequeQueriesBuilder();
                DOMUtils.readChildren(root, SelectionDOMUtils.getQueryElementConsumer(bdfServer.getFichotheque(), fichothequeQueriesBuilder));
                FichothequeQueries fichothequeQueries = fichothequeQueriesBuilder.toFichothequeQueries();
                if (fichothequeQueries.getMotcleQueryList().isEmpty()) {
                    throw BdfErrors.unsupportedParameterValue(InteractionConstants.XML_PARAMNAME, "Missing fiche-query elements");
                }
                for (MotcleQuery mq : fichothequeQueries.getMotcleQueryList()) {
                    builder.add(mq);
                }
        }
    }


    private static class NotFoundBdfInstruction implements BdfInstruction {

        private NotFoundBdfInstruction() {

        }

        @Override
        public short getBdfUserNeed() {
            return BdfInstructionConstants.NONE_BDFUSER;
        }

        @Override
        public ResponseHandler runInstruction(BdfUser bdfUser) {
            return null;
        }

    }

}
