/* FichothequeLib_Xml - Copyright (c) 2008-2022 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package net.fichotheque.xml.extraction;

import java.io.IOException;
import java.util.Locale;
import net.fichotheque.SubsetItem;
import net.fichotheque.SubsetKey;
import net.fichotheque.corpus.Corpus;
import net.fichotheque.corpus.FicheMeta;
import net.fichotheque.corpus.fiche.FicheBlocks;
import net.fichotheque.croisement.Croisement;
import net.fichotheque.croisement.Lien;
import net.fichotheque.extraction.ExtractionConstants;
import net.fichotheque.extraction.def.FicheFilter;
import net.fichotheque.extraction.def.MotcleFilter;
import net.fichotheque.extraction.filterunit.FilterUnit;
import net.fichotheque.extraction.run.FicheExtractInfo;
import net.fichotheque.extraction.run.MotcleExtractInfo;
import net.fichotheque.namespaces.ExtractionSpace;
import net.fichotheque.thesaurus.Motcle;
import net.fichotheque.utils.FichothequeUtils;
import net.mapeadores.util.annotation.Nullable;
import net.mapeadores.util.attr.Attribute;
import net.mapeadores.util.attr.Attributes;
import net.mapeadores.util.exceptions.NestedIOException;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.localisation.LangContext;
import net.mapeadores.util.localisation.ListLangContext;
import net.mapeadores.util.localisation.UserLangContext;
import net.mapeadores.util.text.LabelUtils;
import net.mapeadores.util.xml.AppendableXMLWriter;
import net.mapeadores.util.xml.XMLUtils;
import net.mapeadores.util.xml.XMLWriter;


/**
 *
 * @author Vincent Calame
 */
public final class ExtractionXMLUtils {

    private ExtractionXMLUtils() {
    }

    public static int getMainPoids(Croisement croisement) {
        if (croisement == null) {
            return 0;
        }
        return croisement.getLienList().get(0).getPoids();
    }

    public static void writeCroisement(XMLWriter xmlWriter, Croisement croisement) throws IOException {
        if (croisement == null) {
            return;
        }
        for (Lien lien : croisement.getLienList()) {
            xmlWriter.startOpenTag("lien");
            xmlWriter.addAttribute("mode", lien.getMode());
            xmlWriter.addAttribute("poids", lien.getPoids());
            xmlWriter.closeEmptyTag();
        }
    }

    public static void writePhrase(XMLWriter xmlWriter, SubsetItem subsetItem, String phraseName, LangContext langContext, int extractVersion) throws IOException {
        String tagName;
        if (extractVersion == ExtractionConstants.INITIAL_VERSION) {
            tagName = "intitule";
            xmlWriter.openTag("intitule");
        } else {
            tagName = "phrase";
            xmlWriter.startOpenTag(tagName);
            xmlWriter.addAttribute("name", phraseName);
            xmlWriter.endOpenTag();
        }
        if (langContext instanceof ListLangContext) {
            for (ListLangContext.Unit unit : (ListLangContext) langContext) {
                Lang lang = unit.getLang();
                Locale locale = unit.getFormatLocale();
                String title = FichothequeUtils.getNumberPhrase(subsetItem, phraseName, lang, locale);
                LabelUtils.addLabel(xmlWriter, lang, title);
            }
        } else if (langContext instanceof UserLangContext) {
            UserLangContext userLangContext = (UserLangContext) langContext;
            Lang workingLang = userLangContext.getWorkingLang();
            Locale locale = userLangContext.getFormatLocale();
            String title = FichothequeUtils.getNumberPhrase(subsetItem, phraseName, workingLang, locale);
            LabelUtils.addLabel(xmlWriter, null, title);
        }
        xmlWriter.closeTag(tagName);
    }

    public static String ficheBlocksToExtractionString(FicheBlocks ficheBlocks, ExtractParameters extractParameters, @Nullable SubsetKey defaultCorpusKey) {
        StringBuilder buf = new StringBuilder();
        AppendableXMLWriter xmlWriter = XMLUtils.toXMLWriter(buf, -999);
        try {
            xmlWriter.openTag("section");
            if (ficheBlocks != null) {
                Corpus defaultCorpus = null;
                if (defaultCorpusKey != null) {
                    defaultCorpus = (Corpus) extractParameters.getExtractionContext().getFichotheque().getSubset(defaultCorpusKey);
                }
                FicheBlockXMLPart ficheBlockXMLPart = new FicheBlockXMLPart(xmlWriter, extractParameters);
                ficheBlockXMLPart.setDefaultCorpus(defaultCorpus);
                ficheBlockXMLPart.addFicheBlocks(ficheBlocks);
            }
            xmlWriter.closeTag("section");
        } catch (IOException ioe) {
            throw new NestedIOException(ioe);
        }
        return buf.toString();
    }

    public static MotcleExtractInfo toMotcleExtractInfo(Motcle motcle, MotcleFilter motcleFilter) {
        return new SimpleMotcleExtractInfo(motcle, motcleFilter);
    }

    public static FicheExtractInfo toFicheExtractInfo(FicheMeta ficheMeta, FicheFilter ficheFilter) {
        return new SimpleFicheExtractInfo(ficheMeta, ficheFilter);
    }

    public static boolean hideEmpty(FilterUnit filterUnit, ExtractParameters extractParameters) {
        if (filterUnit != null) {
            return hideEmpty(filterUnit.getAttributes(), extractParameters);
        }
        return !extractParameters.isWithEmpty();
    }

    public static boolean hideEmpty(Attributes attributes, ExtractParameters extractParameters) {
        Attribute attribute = attributes.getAttribute(ExtractionSpace.HIDE_ATTRIBUTEKEY);
        if (attribute != null) {
            switch (attribute.getFirstValue()) {
                case ExtractionSpace.NEVER_HIDE:
                    return false;
                case ExtractionSpace.EMPTY_HIDE:
                    return true;
            }
        }
        return !extractParameters.isWithEmpty();
    }

    public static void addGroupsAttribute(XMLWriter xmlWriter, FilterUnit filterUnit) throws IOException {
        if (filterUnit == null) {
            return;
        }
        addGroupsAttribute(xmlWriter, filterUnit.getAttributes());
    }

    public static void addGroupsAttribute(XMLWriter xmlWriter, Attributes attributes) throws IOException {
        Attribute groups = attributes.getAttribute(ExtractionSpace.GROUPS_ATTRIBUTEKEY);
        if (groups == null) {
            return;
        }
        StringBuilder buf = new StringBuilder();
        for (String value : groups) {
            if (buf.length() > 0) {
                buf.append(' ');
            }
            buf.append('~');
            buf.append(value);
        }
        xmlWriter.addAttribute("groups", buf.toString());
    }


    private static class SimpleMotcleExtractInfo implements MotcleExtractInfo {

        private final Motcle motcle;
        private final MotcleFilter motcleFilter;

        private SimpleMotcleExtractInfo(Motcle motcle, MotcleFilter motcleFilter) {
            this.motcle = motcle;
            this.motcleFilter = motcleFilter;
        }

        @Override
        public Motcle getMotcle() {
            return motcle;
        }

        @Override
        public MotcleFilter getMotcleFilter() {
            return motcleFilter;
        }

        @Override
        public Croisement getCroisement() {
            return null;
        }

    }


    private static class SimpleFicheExtractInfo implements FicheExtractInfo {

        private final FicheFilter ficheFilter;
        private final FicheMeta ficheMeta;

        private SimpleFicheExtractInfo(FicheMeta ficheMeta, FicheFilter ficheFilter) {
            this.ficheMeta = ficheMeta;
            this.ficheFilter = ficheFilter;
        }

        @Override
        public FicheFilter getFicheFilter() {
            return ficheFilter;
        }

        @Override
        public FicheMeta getFicheMeta() {
            return ficheMeta;
        }

        @Override
        public Croisement getCroisement() {
            return null;
        }

        @Override
        public Object getGroupClauseObject() {
            return null;
        }

    }


}
